//coded by Feifei Li, contact: lifeifei@cs.fsu.edu

#include "udb.h"
#include "mtimer.h"
#include <queue>

bool rank_xtuple(XTuple t1, XTuple t2)
{
	return t1.prank<t2.prank;
}

class PrioritizeTask { 
      public : int operator()( const XTuple &x, const XTuple &y ) 
      { return x.prank < y.prank; } 
};

int main(int argc, char **argv)
{
	if (argc!=5)
	{
		cout<<"Usage: "<<argv[0]<<" input_file "<<" m(0 to read all) "<<" k(0 to rank all) "<<" showRes(0,1)"<<endl;
		return -1;
	}
	priority_queue<XTuple, vector<XTuple>, PrioritizeTask> topk;

    
	unsigned int m=atoi(argv[2]);
	unsigned int k=atoi(argv[3]);
	int showRes=atoi(argv[4]);

	
	Udb db(0, argv[1]);
	db.readRecords(m);
	db.getPv();
	if (k==0)
		k=db.N;

	vector<XTuple> xrank;
	MTimer timer;
	double r=0.0;
	double rv=0.0;
	double pv=0.0;
	unsigned long val=0;
	__gnu_cxx::hash_map<unsigned long, double>::iterator it;	

	timer.go();	
	for (unsigned int i=0;i<db.N;i++)
	{
		r=0.0;
		rv=0.0;
		for (int j=0;j<db.xtps[i].xdegree;j++)
		{
			val=(unsigned long)db.xtps[i].tuples[j].value;
			it=db.Pv.find(val);		
			pv=(*it).second;
			r+=db.xtps[i].tuples[j].confidence*(pv-rv);
			rv+=db.xtps[i].tuples[j].confidence;
		}
		db.xtps[i].prank=r;
		//reuse the xtuple class to represent the ranking information
		XTuple xr(i, 0, 0, 0, db.xtps[i].prank);
		topk.push(xr);
		if (topk.size()>k)
			topk.pop();
		//xrank.push_back(xr);
	}	
	//sort(xrank.begin(), xrank.end(),rank_xtuple);
	timer.stop();
	timer.update();

	vector<unsigned int> rankpos;	
	rankpos.resize(k);
	for (unsigned int i=0;i<k;i++)
	{
		rankpos[k-i-1]=topk.top().id;
		topk.pop();
	}
     
	if (showRes==1)
	{
        	db.printXTuples(rankpos);
		//cout<<endl<<"UDB with Top-k= "<<k<<" x-ranked of xtuples "<<endl;
    	}
	else
	{
	  cout<<k<<" "<<db.N<<" "<<db.hist.size()<<" ";
	  timer.printStat();
	  cout<<endl;
	}
    return 1;	
}
