//coded by Feifei Li, contact: lifeifei@cs.fsu.edu

#include "udb.h"
#include "mtimer.h"

bool rank_xtuple(XTuple t1, XTuple t2)
{
	return t1.prank<t2.prank;
}

bool rankL_xtuple(XTuple t1, XTuple t2)
{
	return t1.prankL<t2.prankL;
}

bool rankH_xtuple(XTuple t1, XTuple t2)
{
	return t1.prankH<t2.prankH;
}

void swap(vector<XTuple>& tps, unsigned int a, unsigned int b)
{
  XTuple xtemp(tps[b]);
  tps[b]=tps[a];
  tps[a]=xtemp;
}

unsigned int partition(vector<XTuple>& tps, unsigned int left, unsigned int right, unsigned int pivotIndex)
{
  double pivot=tps[pivotIndex].prank;
  swap(tps, pivotIndex, right);
  unsigned int index=left;
  for (unsigned int i=left;i<right;i++) {
    if (tps[i].prank < pivot) {
      swap(tps, index, i);
      index++;
    }
  }
  swap(tps, right, index);
  return index;
}

double select(vector<XTuple>& tps, unsigned int k, unsigned int left, unsigned int right)
{
  if (left==right)
    return tps[k].prank;
  unsigned int pivotIndex=rand()%(right-left)+left;
  unsigned int newIndex = partition(tps, left, right, pivotIndex);
  if (k==newIndex)
    return tps[k].prank;
  else if (k < newIndex)
    return select(tps, k, left, newIndex-1);
  else
    return select(tps, k, newIndex+1, right);
}

int main(int argc, char **argv)
{
	if (argc!=6)
	{
		cout<<"Usage: "<<argv[0]<<" input_file "<<" m(0 to read all) "<<" k(0 to rank all) "<<" N'-lookahead "<<" showRes(0,1)"<<endl;
		return -1;
	}
    
	unsigned int m=atoi(argv[2]);
	unsigned int k=atoi(argv[3]);
	unsigned int np=atoi(argv[4]);
	int showRes=atoi(argv[5]);

	Udb db(0, argv[1]);
	db.readRecords(m);
	vector<XTuple> xrank;
	MTimer timer;
	srand(time(NULL));
	timer.go();
	unsigned int n=0;
	double rp=0.0, rm=0.0;
	double tmp1=0.0, tmp2=0.0, tmp3=0.0;
	unsigned int lookahead=0;
	while (n<db.N) {
	  rp=0.0;
	  rm=0.0;
	  xrank.clear();
          lookahead=n+np;
          if (lookahead>=db.N)
		lookahead=db.N-1;
	  for (unsigned int i=0;i<=n;i++) {
	    if (i!=n) {
	      db.xtps[i].prankL+=db.xtps[i].rank(db.xtps[n]);
	      db.xtps[n].prankL+=db.xtps[n].rank(db.xtps[i]);
	      tmp1=db.chebyshev(i, n, lookahead);
	    }
	    //tmp=db.xtps[i].markov(db.xtps[n].expvalue);
	    //db.xtps[i].prank+=(db.N-n)*db.xtps[i].markv*db.xtps[n].expvalue;
	    tmp2=db.xtps[i].markov(db.xtps[lookahead].expvalue);
	    tmp3=db.xtps[i].markov(db.xtps[n].expvalue);
	    db.xtps[i].prankH=db.xtps[i].prankL+tmp1+(db.N-lookahead)*tmp2;
	    rm+=tmp3;
	    //rm+=db.xtps[i].markv*db.xtps[n].expvalue;
	    XTuple xr(i, 0, 0, 0, db.xtps[i].prankH, db.xtps[i].prankL, db.xtps[i].prankH);
	    xrank.push_back(xr);
	  }
	  if (n>=k-1) {
	    rp=select(xrank, k-1, 0, n);
	    rm=n-rm;  
	    //cout<<"N "<<db.N<<" n "<<n<<" r+ "<<rp<<" r- "<<rm<<" "<<db.xtps[n].expvalue<<endl;
	    if (rp<=rm)
	      break;
	  }
	  n++;
	}
	
	timer.stop();
	timer.update();
	//reuse the xtuple class to represent the ranking information
	sort(xrank.begin(), xrank.end(), rankL_xtuple);
	vector<unsigned int> rankpos;	
	for (unsigned int i=0;i<=n;i++)
		rankpos.push_back(xrank[i].id);
     
	if (showRes==1)
	{
        	db.printXTuples(rankpos);

    	}
	cout<<endl<<"UDB with Top-k= "<<k<<" x-ranked of xtuples found in the first n="<<n<<" xtuples."<<endl;
	//cout<<select(db.xtps, k-1, 0, db.xtps.size()-1)<<endl;
	timer.printStat();
    return 1;	
}
