//coded by Feifei Li, contact: lifeifei@cs.fsu.edu

#include "udb.h"
#include "mtimer.h"
#include <cmath>

bool rank_xtuple(XTuple t1, XTuple t2)
{
	return t1.prank<t2.prank;
}

bool rankL_xtuple(XTuple t1, XTuple t2)
{
	return t1.prankL<t2.prankL;
}

bool rankH_xtuple(XTuple t1, XTuple t2)
{
	return t1.prankH<t2.prankH;
}

void swap(vector<XTuple>& tps, unsigned int a, unsigned int b)
{
  XTuple xtemp(tps[b]);
  tps[b]=tps[a];
  tps[a]=xtemp;
}

unsigned int partition(vector<XTuple>& tps, unsigned int left, unsigned int right, unsigned int pivotIndex)
{
  double pivot=tps[pivotIndex].prank;
  swap(tps, pivotIndex, right);
  unsigned int index=left;
  for (unsigned int i=left;i<right;i++) {
    if (tps[i].prank < pivot) {
      swap(tps, index, i);
      index++;
    }
  }
  swap(tps, right, index);
  return index;
}

double select(vector<XTuple>& tps, unsigned int k, unsigned int left, unsigned int right)
{
  if (left==right)
    return tps[k].prank;
  unsigned int pivotIndex=rand()%(right-left)+left;
  unsigned int newIndex = partition(tps, left, right, pivotIndex);
  if (k==newIndex)
    return tps[k].prank;
  else if (k < newIndex)
    return select(tps, k, left, newIndex-1);
  else
    return select(tps, k, newIndex+1, right);
}

int main(int argc, char **argv)
{
	if (argc!=7)
	{
		cout<<"Usage: "<<argv[0]<<" input_file "<<" m(0 to read all) "<<" k(0 to rank all) "<<" showRes(0,1)"<<" k'(0 is n) "<<" exact_res"<<endl;
		cerr<<"test"<<endl;
		return -1;
	}
    
	unsigned int m=atoi(argv[2]);
	unsigned int k=atoi(argv[3]);
	int showRes=atoi(argv[4]);
	unsigned int kp=atoi(argv[5]);


	Udb db(0,  argv[1]);
	db.readRecords(m);
	vector<XTuple> xrank;
	MTimer timer;
	srand(time(NULL));
	timer.go();
	unsigned int n=0;
	double rp=0.0, rm=0.0;
	double tmp=0.0;
	while (n<db.N) {
	  rp=0.0;
	  rm=0.0;
	  xrank.clear();
	  for (unsigned int i=0;i<=n;i++) {
	    if (i!=n) {
	      db.xtps[i].prankL+=db.xtps[i].rank(db.xtps[n]);
	      db.xtps[n].prankL+=db.xtps[n].rank(db.xtps[i]);
	    }
	    tmp=db.xtps[i].markov(db.xtps[n].expvalue);
	    //db.xtps[i].prank+=(db.N-n)*db.xtps[i].markv*db.xtps[n].expvalue;
	    db.xtps[i].prankH=db.xtps[i].prankL+(db.N-n)*tmp;
	    rm+=tmp;
	    //rm+=db.xtps[i].markv*db.xtps[n].expvalue;
	    XTuple xr(i, 0, 0, 0, db.xtps[i].prankH, db.xtps[i].prankL, db.xtps[i].prankH);
	    xrank.push_back(xr);
	  }
	  if (n>=k-1) {
	    rp=select(xrank, k-1, 0, n);
	    rm=n-rm;  
	    //cout<<"N "<<db.N<<" n "<<n<<" r+ "<<rp<<" r- "<<rm<<" "<<db.xtps[n].expvalue<<endl;
	    if (rp<=rm)
	      break;
	  }
	  n++;
	}
	
	timer.stop();
	timer.update();
	//reuse the xtuple class to represent the ranking information
	sort(xrank.begin(), xrank.end(), rankL_xtuple);
	vector<unsigned int> rankpos;	
	for (unsigned int i=0;i<kp;i++)
		rankpos.push_back(xrank[i].id);
     
	if (showRes==1)
	{
        	//db.printXTuples(rankpos);
	  __gnu_cxx::hash_map<unsigned int, double> approxR;
	  __gnu_cxx::hash_map<unsigned int, double> exactR;
	  __gnu_cxx::hash_map<unsigned int, double> exactn;
	  __gnu_cxx::hash_map<unsigned int, double> approxn;
	  for (unsigned int i=0;i<kp;i++)
	    approxR.insert(pair<unsigned int, double>(db.xtps[xrank[i].id].id, xrank[i].prankL));
	  for (unsigned int i=0;i<n;i++)
	    approxn.insert(pair<unsigned int, double>(db.xtps[xrank[i].id].id, xrank[i].prankL));
	  std::ifstream fin;
	  fin.open(argv[6]);                   
	  if (!fin)
	       std::cerr << "Cannot open file " << argv[6] << std::endl;        
	  unsigned int id;
	  double rank;
	  for (unsigned int i=0;i<k;i++)
	  {
	    fin >> id >> rank;
	    exactR.insert(pair<unsigned int, double>(id, rank));
	    exactn.insert(pair<unsigned int, double>(id, rank));
	  }
	  for (unsigned int i=k;i<db.N;i++)
	  {
	    fin >> id >> rank;
	    exactn.insert(pair<unsigned int, double>(id, rank));
	  }
	  fin.close();

	  double precision=0.0, recall=0.0;
	  double rankErrorR=0.0, rankError=0.0;
	  double krankErrorR=0.0, krankError=0.0;
	  __gnu_cxx::hash_map<unsigned int, double>::iterator itA;
	  __gnu_cxx::hash_map<unsigned int, double>::iterator itE;
	  itE=exactR.begin();
	  for (unsigned int i=0;i<k;i++)
	  {
	    id=(*itE).first;
	    itA=approxR.find(id);
	    if (itA!=approxR.end())
	    {
	      recall+=1.0;
	    }
	    itE++;
	  }

	  itA=approxR.begin();
	  for (unsigned int i=0;i<k;i++)
	  {
	    id=(*itA).first;
	    itE=exactR.find(id);
	    if (itE!=exactR.end())
	    {
	      precision+=1.0;
	    }
	    itA++;
	  }

	  double found=0.0;
	  itA=approxn.begin();
	  for (unsigned int i=0;i<n;i++)
	  {
	    id=(*itA).first;
	    itE=exactn.find(id);
	    if (itE!=exactn.end())
	    {
	      found+=1.0;
	      double erank=(*itE).second;
	      double arank=(*itA).second;
	      rankErrorR+=abs(erank-arank)/erank;
	      rankError+=abs(erank-arank);
	    }
	    itA++;
	  }
	  rankErrorR=rankErrorR/found;
	  rankError=rankError/found;

	  found=0.0;
	  itA=approxR.begin();
	  for (unsigned int i=0;i<k;i++)
	  {
	    id=(*itA).first;
	    itE=exactn.find(id);
	    if (itE!=exactn.end())
	    {
	      found+=1.0;
	      double erank=(*itE).second;
	      double arank=(*itA).second;
	      krankErrorR+=abs(erank-arank)/erank;
	      krankError+=abs(erank-arank);
	    }
	    itA++;
	  }
	  krankErrorR=krankErrorR/found;
	  krankError=krankError/found;

	  precision=precision/(double)kp;
	  recall=recall/(double)k;
	  double pruned=1.0-(double)n/(double)db.N;
	  cerr<<k<<" "<<db.N<<" "<<n<<" "<<pruned<<" "<<kp<<" "<<recall<<" "<<precision<<" "<<krankError<<" "<<krankErrorR<<" "<<rankError<<" "<<rankErrorR<<endl;
    	}
	//cout<<endl<<"UDB with Top-k= "<<k<<" x-ranked of xtuples found in the first n="<<n<<" xtuples."<<endl;
	//cout<<select(db.xtps, k-1, 0, db.xtps.size()-1)<<endl;

	  double pruned=1.0-(double)n/(double)db.N;
	  cout<<k<<" "<<db.N<<" "<<n<<" "<<pruned<<" "<<db.hist.size()<<" ";
	  timer.printStat();
	  cout<<endl;

    return 1;	
}
