//coded by Feifei Li, contact: lifeifei@cs.fsu.edu

#include "udb.h"

bool rank_tuple(Tuple t1, Tuple t2)
{
	return t1.value>t2.value;
}


Udb::Udb(unsigned int N,   char* file_input)
{
   this->N=N;
   fin.open(file_input);                   
   if (!fin)
    	std::cerr << "Cannot open file " << file_input << std::endl;        
}

Udb::~Udb()
{
   if (fin.is_open())
      fin.close(); 
   for (int i=0;i<N;i++)
	xtps[i].tuples.clear();
   xtps.clear();
}

bool Udb::readNext()
{
    double expvalue;
    double var;
    double value;
    double did=0;
    double confidence;
    unsigned int id=0;
    int xdegree;

  if (!fin)
    return false;
   fin >> did >> xdegree >> expvalue >> var;
   id=(unsigned int)did;
   XTuple xtp(id,xdegree, expvalue, var, 0.0);
   double r=0.0;
   for (int i=0;i<xdegree;i++)
   {	
 	 fin >> value >> confidence;
	 r=r+confidence/value;
	 Tuple tp(value, confidence);
	 xtp.tuples.push_back(tp);
	 tps.push_back(tp);
   }
   xtp.markv=r;
   xtps.push_back(xtp);
   N++;	      
}

bool Udb::readRecords(unsigned int m)
{ 
    if (!fin)
       return false;
    unsigned int counter=0;
    double expvalue;
    double var;
    double value;
    double did=0;
    double confidence;
    int xdegree;
    unsigned int id=0;
    __gnu_cxx::hash_map<unsigned long, unsigned int>::iterator it;	

    while(fin)
    {
	      if (!fin.good()) continue; //skip newlines, etc
              fin >> did >> xdegree >> expvalue >> var;
	      id=(unsigned int)did;
	      XTuple xtp(id,xdegree, expvalue, var, 0.0);
	      double r=0.0;
	      for (int i=0;i<xdegree;i++)
	      {	
 		     fin >> value >> confidence;
		     //sanity check, in our data file, confidence is guaranteed to be between 0 and 1; and value is positive
		      //else if (value>1000000)
			//value=(unsigned long)value%1000000;
		     it=hist.find((unsigned long)value);
		     if (it==hist.end())
		     {
		       hist.insert(pair<unsigned long, unsigned int>((unsigned long)value, 1));
		     }
		     else
		       (*it).second+=1;
		     r=r+confidence/value;
		     Tuple tp(value, confidence);
	     	     xtp.tuples.push_back(tp);
		     tps.push_back(tp);
	      }
	      sort(xtp.tuples.begin(), xtp.tuples.end(), rank_tuple);
	      xtp.markv=r;
	      xtps.push_back(xtp);
	      counter++;
	      if (m!=0)
	      {
		      if (counter>=m)
			break;
	      }
   }
   //remove the weird bug that C++ has to almost read the last line twice
   if (m==0)
   {
      counter--;
      xtps.pop_back();
   }
   N=counter;
   return true;  
}

double Udb::chebyshev(unsigned int i, unsigned int n, unsigned int lookahead)
{
	double r=0.0, tmp=0.0;
	if (xtps[i].chevFirst==true)
	{
		for (unsigned int j=n;j<=lookahead;j++)
		{
			for (int s=0;s<xtps[i].xdegree;s++)
			{
				if (xtps[i].tuples[s].value <= xtps[j].expvalue)
					r+=xtps[i].tuples[s].confidence;
				else
				{
					tmp=xtps[j].var/pow(xtps[i].tuples[s].value-xtps[j].expvalue,2.0);
					if (tmp>1.0)
						tmp=1.0;
					r+=xtps[i].tuples[s].confidence*tmp;	
				}
			}	
		}
		xtps[i].chev=r;
		xtps[i].chevFirst=false;
	}
	else
	{
		for (int s=0;s<xtps[i].xdegree;s++)
		{
			if (xtps[i].tuples[s].value <= xtps[n-1].expvalue)
				xtps[i].chev=xtps[i].chev-xtps[i].tuples[s].confidence;
			else
			{
				tmp=xtps[n-1].var/pow(xtps[i].tuples[s].value-xtps[n-1].expvalue,2.0);
				if (tmp>1.0)
					tmp=1.0;
				xtps[i].chev=xtps[i].chev-xtps[i].tuples[s].confidence*tmp;	
			}
			if (xtps[i].tuples[s].value <= xtps[lookahead].expvalue)
				xtps[i].chev=xtps[i].chev+xtps[i].tuples[s].confidence;
			else
			{
				tmp=xtps[lookahead].var/pow(xtps[i].tuples[s].value-xtps[lookahead].expvalue,2.0);
				if (tmp>1.0)
					tmp=1.0;
				xtps[i].chev=xtps[i].chev+xtps[i].tuples[s].confidence*tmp;	
			}
		}
	}
	return xtps[i].chev; 
}

void Udb::printDB()
{
     cout<<"xtp-id xtp-degree expvalue var x-rank x-rankL x-rankH (value prob):xdegree pairs"<<endl<<endl;
     for (unsigned int i=0;i<N;i++)
         cout<<xtps[i];     
     cout<<endl<<"xtp-id xtp-degree expvalue var x-rank x-rankL x-rankH (value prob):xdegree pairs"<<endl;
}

void Udb::printXTuples(vector<unsigned int> rankpos)
{
	//cout<<"xtp-id xtp-degree expvalue var x-rank x-rankL x-rankH (value prob):xdegree pairs"<<endl<<endl;
	for (unsigned int i=0;i<rankpos.size();i++)
		cout<<xtps[rankpos[i]];
	//cout<<endl<<"xtp-id xtp-degree expvalue var x-rank x-rankL x-rankH (value prob):xdegree pairs"<<endl;
}

void Udb::getPv()
{
	sort(tps.begin(),tps.end(),rank_tuple);
	__gnu_cxx::hash_map<unsigned long, double>::iterator it;	
	unsigned long val=0;
	unsigned int U=tps.size();
	double r=0.0;
	unsigned long lastV=0;  
	for (unsigned int i=0;i<U;i++)
	{
		val=(unsigned long)tps[i].value;
		if (val!=lastV)
			Pv.insert(pair<unsigned long, double>(val, r));
		r+=tps[i].confidence;
		lastV=val;	
		//it=Pv.find(val);
		//if (it==Pv.end())
		//	Pv.insert(pair<unsigned long, double>(val, r));
		//r+=tps[i].confidence;
	}	
}
