//coded by Feifei Li, contact: lifeifei@cs.fsu.edu

#include "medianrank.h"

MedianRank::MedianRank(int k)
{
   this->k=k;    
   this->db=NULL;
   elapsed=0.0;                       
}

MedianRank::~MedianRank()
{
   if (db!=NULL)
      delete db;                                                   
}

bool MedianRank::initDB(unsigned int N, char* input_file, char* xtuple_file, int xdegree, unsigned int numxtuple)
{
   db= new Udb(N, input_file);
   if (!(db->fin))
      return false;
   db->readRecords();
   db->buildXtuples(xtuple_file, xdegree, numxtuple);
   //db->sortByScore(); not necessary if the input db has already sorted, which is the case for our generator.
   return true;     
}

bool MedianRank::has_preceeding_alternative(unsigned int tp_sortid)
{
     bool have=true;
     unsigned int x_sortid;
     if (db->xtuples[tp_sortid].size()==0 )
        have=false;
     else
     {
         /*for (int j=0;j<db->xdegree-1;j++)
         {
             xid=db->xtuples[tp_sortid][j];
             if (xid<tp_sortid)
             {
                none=false;
                break;
             }
         }
         none=true;*/
         //as xid list is sorted, check the first one is good enough;
         x_sortid=db->xtuples[tp_sortid][0];
         if (x_sortid<tp_sortid)
            have=true;
         else
            have=false;
     }
     return have;
}

void MedianRank::findTopk()
{
   unsigned int i=0;
   unsigned int index;
   unsigned int j=0;
   bool stop=false;
   double pprime=0.0;
   unsigned int barpi_loc;
   vector<double> rij_minus;
   __gnu_cxx::hash_map<unsigned int, unsigned int, GnuHash>::iterator it;	      
   
   rij_minus.resize(this->k+1);
   Rij.resize(this->k+1);
   topk.resize(this->k+1);
   krank.resize(this->k+1);
   for (i=0;i<(unsigned int)k+1;i++)
   {
       Rij[i]=0.0;
       krank[i]=0.0;
   }    
   Rij[0]=1.0;
   i=1;
   t1 = times(&st);
   gettimeofday (&start,NULL);
   while (true)
   {
          Tuple tp=db->tuples[i-1];
          index=k;
          if (i<(unsigned int)k)
              index=i;
              
         if (has_preceeding_alternative(tp.sortid)==false)
         {     
               //no preceeding alternatives
               barp.push_back(tp.confidence);
               barp_location.insert(pair<unsigned int, unsigned int>(tp.xid, barp.size()-1)); 
               //using xtuple id (instead of sortid) to retrieve barpi's location(in barDi)
               for (j=1;j<index+1;j++)
               {
                     double newconf=tp.confidence*Rij[j-1];   
                     if (krank[j]<newconf)
                     {
                          krank[j]=newconf;
                          topk[j]=tp;
                     }      
               }
               for (j=index;j>=1;j--)
                   Rij[j]=tp.confidence*Rij[j-1]+(1-tp.confidence)*Rij[j];
               Rij[0]=(1-tp.confidence)*Rij[0];
         }
         else
         {
             //have preceeding alternatives
             it=barp_location.find(tp.xid);
             barpi_loc=(*it).second;
             pprime=barp[barpi_loc];
             barp[barpi_loc]=0.0;
             computeRij(barp,rij_minus);
             for (j=1;j<index+1;j++)
             {
                  double newconf=tp.confidence*rij_minus[j-1];   
                  if (krank[j]<newconf)
                  {
                       krank[j]=newconf;
                       topk[j]=tp;
                  }      
             }
             barp[barpi_loc]=pprime+tp.confidence;
             computeRij(barp,Rij);
         }
         
         if (i>=(unsigned int)k)
         {
             stop=checkTerminate(krank, Rij);
             if (stop)
             {
                 this->n=i;
                  break;
              }
          }  
          i++;
          if (i>db->N)
          {
            this->n=db->N;
            break;
          }
   }
   gettimeofday (&end,NULL);
   t2 = times(&ed);
   usertime=((double)(ed.tms_utime-st.tms_utime))/(HZ);
   systemtime=((double)(ed.tms_stime-st.tms_stime))/(HZ);
   realtime=((double)(t2-t1))/(HZ);
   //db->timeval_subtract(&tdiff, &end, &start);
   elapsed=tdiff.tv_sec+tdiff.tv_usec/1000000.0;
   getMemsize();
}

void MedianRank::printStatistics()
{	
    
    cout<<endl<<"scan depth: "<<this->n<<endl<<endl;
    printf("user time=%f \n", usertime);
    printf("system time=%f \n", systemtime);
    printf("real time=%f \n", realtime);
    printf("clock time=%f \n", elapsed); 
    cout<<"max mem usage(in bytes)= "<<memusage<<endl;
    cout<<"max mem usage(in KB)= "<<(double)memusage/1024.0<<endl;
    cout<<"max mem usage(in MB)= "<<(double)memusage/(1024.0*1024.0)<<endl;
}


bool MedianRank::checkTerminate(vector<double> krank, vector<double>Rij)
{
     int i=0;
     bool stop=true;
     double maxrnl=0.0;
     for (i=1;i<=this->k;i++)
     {
         if (maxrnl<Rij[i-1])
              maxrnl=Rij[i-1];
         if (krank[i]<maxrnl)
         {
              stop=false;
              break;                      
          }  
     }
     return stop;
}

void MedianRank::printTopk()
{
     int i=0;
     for (i=1;i<=k;i++)
     {
         Tuple tp=topk[i];
         cout<<tp<<" "<<krank[i]<<endl;
     }    
}

void MedianRank::computeRij(vector<double> pi, vector<double>& rij)
{
     unsigned int size=pi.size();
     unsigned int index;
     unsigned int i, j;
     double p;
     
      for (i=0;i<(unsigned int)k+1;i++)
           rij[i]=0.0;
      rij[0]=1.0;
      
     for (unsigned int i=1;i<=size;i++)
     {
          p=pi[i-1];
          index=k;
          if (i<(unsigned int)k)
              index=i;
          for (j=index;j>=1;j--)
              rij[j]=p*rij[j-1]+(1-p)*rij[j];
          rij[0]=(1-p)*rij[0];
     }     
}

void MedianRank::test_hashmap(__gnu_cxx::hash_map<unsigned int, unsigned int, GnuHash> mp)
{
     mp.insert(pair<unsigned int, unsigned int>(10, 2));
     mp.insert(pair<unsigned int, unsigned int>(3, 5));
     mp.insert(pair<unsigned int, unsigned int>(2, 4));
     
     __gnu_cxx::hash_map<unsigned int, unsigned int, GnuHash>::iterator it;	      
     it=mp.begin();
     cout<<(*it).first<<" "<<(*it).second<<endl; 
     it++;
          cout<<(*it).first<<" "<<(*it).second<<endl;
          it++;
               cout<<(*it).first<<" "<<(*it).second<<endl;
     
     vector<double> p1;
     vector<double> p2;
     p2.resize(3);
     
     p1.push_back(0.4);
     p1.push_back(0.3);
     computeRij(p1, p2);
     cout<<"rij results: "<<endl;
     cout<<p2[0]<<endl;
     cout<<p2[1]<<endl;
     cout<<p2[2]<<endl;
}

void MedianRank::getMemsize()
{
     //dominated by barp_loc, essentially the number of xtuples (no alternative consider as one xtuple and alternatives are combined into one xtuple)    
     memusage=barp_location.size()*Tuple::size(true);   
     memusage+=k*(8+8);
     memusage+=k*Tuple::size(true);
}
