//coded by Feifei Li, contact: lifeifei@cs.fsu.edu

#include "udb.h"
#include "mtimer.h"
#include <queue>

/*
bool rank_xtuple(XTuple t1, XTuple t2)
{
	return t1.prank<t2.prank;
}
*/

class PrioritizeTask { 
      public : int operator()( const Tuple &x, const Tuple &y ) 
      { return x.prank < y.prank; } 
};

int main(int argc, char **argv)
{
	if (argc!=8)
	{
		cout<<"Usage: "<<argv[0]<<" input_file "<<" xid_file "<<" m(0 to read all) "<<" numxtuple "<< " xdegree "<<" k(0 to rank all) "<<" showRes(0,1)"<<endl;
		return -1;
	}
	
    
	unsigned int m=atoi(argv[3]);
	unsigned int numxtuple=atoi(argv[4]);
	int xdegree =atoi(argv[5]);
	unsigned int k=atoi(argv[6]);
	int showRes=atoi(argv[7]);

	
	Udb db(m, argv[1]);
	db.readRecords();
	db.buildXtuples(argv[2], xdegree, numxtuple);

	if (k==0)
		k=db.N;

	MTimer timer;
	double rv=0.0;
	double qi=0.0;
	double rm=0.0;
	double maxalter=0.0;
	unsigned int n=0;
	priority_queue<Tuple, vector<Tuple>, PrioritizeTask> topk;
	//__gnu_cxx::hash_map<unsigned long, double>::iterator it;	

	timer.go();	
	for (unsigned int i=0;i<db.N;i++)
	{	
		rv=0.0;
		for (int j=0;j<db.xtuples[i].size();j++)
		{
			if (db.xtuples[i][j]<i)
				rv+=db.tuples[db.xtuples[i][j]].confidence;
		}
		if (rv>maxalter)
			maxalter=rv;
		db.tuples[i].prank=(qi-rv)/db.tuples[i].confidence;
		topk.push(db.tuples[i]);
		if (topk.size()>k)
			topk.pop();
		qi+=db.tuples[i].confidence;
		rm=qi-maxalter;	
		if (topk.size()>=k && topk.top().prank<rm)
		{	
			n=i+1;
			break;
		}
	}
	timer.stop();
	timer.update();

     
	if (showRes==1)
	{
		vector<Tuple> res;
		res.resize(k);
		for (unsigned int j=0;j<k;j++)
		{	
			res[k-j-1]=topk.top();
			topk.pop();
		}
		for (unsigned int j=0;j<k;j++)
			cout<<res[j]<<endl;

		cout<<endl<<"UDB with Top-k= "<<k<<" ranked of tuples "<<endl;
		cout<<"Pruning terminates after seeing "<<n<<" tuples "<<endl;
    	}
	timer.printStat();
    return 1;	
}
