//coded by Feifei Li, contact: lifeifei@cs.bu.edu

#include "udb.h"

bool compare_tuple_score(Tuple t1, Tuple t2)
{
    return t1.score>=t2.score;       
} 


Udb::Udb(unsigned int N, char* file_input)
{
   this->N=N;
   ts=0.0;
   fin.open(file_input);                   
   if (!fin)
    	std::cerr << "Cannot open file " << file_input << std::endl;        
}

Udb::~Udb()
{
   if (fin.is_open())
      fin.close(); 
}

bool Udb::readRecords()
{ 
    if (!fin)
       return false;
    unsigned int counter=0;
    double score;
    unsigned int id;
    double confidence;
    double r=0.0;
    __gnu_cxx::hash_map<unsigned long, unsigned int>::iterator it;	

    while(fin)
    {
              fin >> id >> score >> confidence;
              /*if (confidence<0.0)
                 confidence=(1+confidence)*0.01;
              else if (confidence>1.0)
                 confidence=0.95;
              if (score<=0)
                 score=1;*/
              if (!fin.good()) continue; //skip newlines, etc
	      it=hist.find((unsigned long)score);
	      if (it==hist.end())
	      {
		  hist.insert(pair<unsigned long, unsigned int>((unsigned long)score, 1));
	      }
	      else
		 (*it).second+=1;
              Tuple tp(counter, id, score, confidence);
              tuples.push_back(tp);
	      r+=confidence;
              counter++;
	      if (N!=0)
	      {
                 if (counter>=N)
                    break;
	      }
    }           
    if (N==0)
	N=counter;
    this->ts=r;
    return true;  
}

void Udb::sortByScore()
{
     sort(tuples.begin(),tuples.end(),compare_tuple_score);     
}

void Udb::printDB()
{
     for (unsigned int i=0;i<N;i++)
         //tuples[i].show();
         cout<<tuples[i]<<endl;     
}

unsigned int Udb::buildXtuples(char* file_input, int xdegree, unsigned int numxtuple)
{
   std::ifstream xfin;
   xfin.open(file_input);                   
   if (!xfin)
   {
    	std::cerr << "Cannot open file " << file_input << std::endl;   
    	this->numxtuple=0;
    	this->xdegree=0;
    	return 0;
   }
   xtuples.resize(this->N);	
  	unsigned int counter=0;
  	int i=0,j=0;
  	unsigned int xtpid[xdegree];
  	
    while(xfin)
    {
              for (i=0;i<xdegree;i++)
              {
                  xfin >> xtpid[i];
              }
              if (!xfin.good()) continue; //skip newlines, etc
              for (i=0;i<xdegree;i++)
              {
                  for (j=0;j<xdegree;j++)
                  {
                      if (i!=j)
                      {
                         xtuples[xtpid[i]-1].push_back(xtpid[j]-1);
                      }
                  }
              }
              counter++;
    }
    if (counter!=numxtuple)
       cout<<"wrong... number of xtuples no equal to the specified value in the file"<<endl;
    this->numxtuple=numxtuple;
    this->xdegree=xdegree;
    return counter;
}

void Udb::printXtuple(unsigned int id)
{
     //only call this function after buildXtuple
     if (xtuples[id].size()!=0)
     {

     	cout<<id<<" its alternatives are: ";
        for (int i=0;i<xdegree-1;i++)
             cout<< tuples[xtuples[id][i]]<<" ";
        cout<<endl;
     }
     else
	cout<<id<<endl;
}

void Udb::printXtuple()
{
	for (unsigned int i=0;i<N;i++)
		printXtuple(i);
}

void Udb::buildqi()
{
	double r=0.0;
	for (unsigned int i=0;i<tuples.size();i++)
	{
		qi.push_back(r);
		r+=tuples[i].confidence;
	}
	this->ts=r;
}


void Udb::getRankDist() {

/*
        unsigned int N;
        unsigned int numxtuple;
        int xdegree;
        std::ifstream fin;
        vector<Tuple> tuples;
        vector< vector<unsigned int> > xtuples;
        vector<unsigned int> preceding_xtuples; // xtuples preceding a current tuple, contains xid's
        vector<double> merged_tuples;  // indexable by xid
        vector<double> qi;
        __gnu_cxx::hash_map<unsigned long, unsigned int> hist;
        double ts;
*/

   vector<double> Rij;
   Rij.resize( numxtuple );
   double prev_Rij;
   unsigned int row_size;

   merged_tuples.resize( numxtuple );
   preceding_xtuples.resize( 0 );
   for( unsigned int l = 0; l < merged_tuples.size(); ++l ) {
      merged_tuples[l] = -1;
   }

   for( unsigned int n = 0; n < N; ++n ) { // the current tuple
      tuples[n].Rj.resize( N );
      for( unsigned int m = 0; m < N; ++m ) {
         tuples[n].Rj[m] = 0;
      }

      Rij[0] = 1;
      row_size = 0;
      for( unsigned int i = 0; i < preceding_xtuples.size(); ++i ) {
         if( preceding_xtuples[i] == tuples[n].xid ) continue;
         for( unsigned int j = 0; j <= row_size; ++j ) {
            if( j == 0 ) {
               prev_Rij = Rij[j];
               Rij[j] = ( 1 - merged_tuples[preceding_xtuples[i]] ) * Rij[j];
            }
            else if( j == row_size ) {
               Rij[j] = merged_tuples[preceding_xtuples[i]] * Rij[j];
            }
            else {
               unsigned int temp = Rij[j];
               Rij[j] = ( prev_Rij * merged_tuples[preceding_xtuples[i]] ) + ( ( 1 - merged_tuples[preceding_xtuples[i]] ) * Rij[j] );
               prev_Rij = temp;
            }
         }           
         row_size++;
      }

      for( unsigned int i = 0; i <= row_size; ++i ) {
         tuples[n].Rj[i] = Rij[i] * tuples[n].confidence;
      }

      if( merged_tuples[tuples[n].xid] < 0 ) {
         preceding_xtuples.push_back( tuples[n].xid );
         merged_tuples[tuples[n].xid] = tuples[n].confidence;
      }
      else {
         merged_tuples[tuples[n].xid] += tuples[n].confidence;
      }
   }
}
                
void Udb::printRankDist() {
   for( unsigned int i = 0; i < tuples.size(); ++i ) {
      cout << tuples[i] << ": ";
      for( unsigned int j = 0; tuples[i].Rj[j] != 0; ++j ) {
         cout << tuples[i].Rj[j] << ' ';
      }
      cout << endl;
   }
}
